// Copyright (c) 2014, Kirth Gersen / kirthalleg@gmail.com.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'dart:io';
import 'dart:typed_data';
import 'dart:isolate';
import 'dart:async';
import 'web/speedtest.dart';
import 'package:args/args.dart';

// todo start server with Isolate.spawnUri to autouptodate code
// this main here should only be a launcher for the isolate
ReceivePort receivePort = null;
SendPort sendPort = null;

// notify parent that we're going down
void shutdown(String cause, { String code:'SHUTDOWN'})
{
  if (sendPort!=null)
  {
    sendPort.send('$code:$cause');
  }
  // then do nothing else, we should exit
}

HttpServer TheServer = null;

// client will send a list of supported protocol
String protocolSelector(List<String> p)
{
  if (p==null) return null; 
  print('$p');
  return p.first;
}

void main(List<String> args, SendPort p) {
  // test if standalone or isolate
  if (p != null)
  {
    receivePort = new ReceivePort();
    sendPort = p;
    sendPort.send(receivePort.sendPort);
  }

  ArgParser parser = new ArgParser();

  int port = 8192; //todo: add defaut port value in a global const 
  parser.addOption('port', abbr: 'p', help:'server port');

  InternetAddress addr = InternetAddress.ANY_IP_V6;// this means V6 and V4 allowed
  parser.addFlag('ipv4', abbr: '4', help:'IPv4 only');

  Duration timeout = new Duration(seconds:120);
  parser.addOption('timeout', abbr: 't', help:'timeout');
  
  ArgResults results;
  try
  { 
    results = parser.parse(args); 
  }
  catch (e)
  {
    print(e.toString());
    print(parser.getUsage());
    shutdown('args error');
    return;
  }
  
  if (results.wasParsed('port'))
  {
    port = int.parse(results['port']);
    if (port < 0 || port > 65535)
    {
      print("error invalid port");
      shutdown('args error');
      return;
    }
  } 
  
  if (results.wasParsed('ipv4'))
  {
    if (results['ipv4'])
    {
      addr = InternetAddress.ANY_IP_V4; // V4 only, no V6
    }
  }
  
  if (results.wasParsed('timeout'))
  {
    int t = int.parse(results['timeout'], onError : (s)=>-1);
    if (t==-1)
    {
      print("error invalid timeout");
      shutdown('args error');
      return;
    }
    timeout = new Duration(seconds:t);
  }
  // init buffer
  TestBuffer.InitBuffer();

  // init server
  HttpServer.bind(addr, port).then((HttpServer server) {
    TheServer = server;
    // todo: investigate odd ipv6 localhost
    print("Speedtest server v$SpeedTestVersion listening on ${addr.address}:$port, timeout is ${timeout.inSeconds} seconds");
    server.serverHeader = "Dart Speed Test ($SpeedTestVersion) by Kirth Gersen";
    server.listen((HttpRequest request) {
      
      //todo: this is the main handler, do a function and catch all errors too
      //todo: spawn a child to handle that request (//isme)
      
      String message = "http connection from: "+ request.connectionInfo.remoteAddress.host.toString();
      try 
      { // catch invalid URIs
        Uri uri = request.requestedUri;
        print('$message. requestedUri = '+uri.toString());
      }
      catch (e)
      {
        print('$message. invalid uri ($e)');
        serveRequest(request, code : HttpStatus.BAD_REQUEST, reason : 'invalid uri!');
        return;
      }
      
      if (WebSocketTransformer.isUpgradeRequest(request))
      {
        // todo: try/catch this
        try
        {
          String clientversion = request.uri.pathSegments[1];
          if (clientversion != SpeedTestVersion)
          {
            // send client en error and assume we need to upgrade the server
            print('client mismatch: $clientversion (our is $SpeedTestVersion)');
            serveRequest(request, code : HttpStatus.NOT_ACCEPTABLE, reason : 'version mismatch');
            UpgradeServer(clientversion);
          }
          else
          {
            SpeedtestServer s = new SpeedtestServer(request.connectionInfo, timeout);
            WebSocketTransformer.upgrade(request, protocolSelector : protocolSelector, compression: CompressionOptions.OFF).then(s.handleWebSocket);
          }
        }
        catch (e)
        {
          print('error: $e');
          serveRequest(request, code : HttpStatus.BAD_REQUEST, reason : 'something went wrong!');
        }
      }
      else {
        print("Regular ${request.method} request for: ${request.uri.path}");
        serveRequest(request);
      }
    });
  });
}

// todo: socket.add can crash the server
// todo: share more code with the client
// todo: this shoud be a Connection ?
class SpeedtestServer
{
  WebSocket _socket;
  HttpConnectionInfo _conninfo;
  List<int> _b = new List<int>(0);
  Duration _timeout;
  Timer _timer;
  int _totalbytes;
  
  SpeedtestServer(this._conninfo, this._timeout) 
  {
    _totalbytes = 0;
  }
  void handleWebSocket(WebSocket socket)
  {
    _socket = socket;
    //print('Client connected! ');
    _socket.listen(onData, onError: onError, onDone: onDone);
    _socket.add('A'+_conninfo.remoteAddress.address);
    _socket.add('V$SpeedTestVersion');
    _timer = new Timer(_timeout,_onTimeout);
  }
  
  void _onTimeout()
  {
    // game over dude, you took too long
    print('timeout!');
    _socket.add("0"); // tell the client we timeout'ed
    _socket.close(WebSocketStatus.POLICY_VIOLATION,'timeout');
  }
  
  void onError(Error e)
  {
    print('error : '+e.toString());
  }
  
  void onData(dynamic s) {
    //print (s.runtimeType.toString());
    if (s is String)
      onCommand(s);
    else
    {
      // we received some binary data  
      Uint8List b = s; // todo: this is not future proof
      // we send back the number of bytes we received
      _socket.add('U' + b.length.toString());
    }
  }
  
  void onCommand(String s)
  {
    {
      if (s=='0')// disconnect requested
      {
        _socket.close(WebSocketStatus.NORMAL_CLOSURE,'client requested');
        return;
      }
      else
      {
        //print('Client sent: $s');
        // parse message here
        int size = int.parse(s, onError:(source)=>1); //int.parse(s,onError:(source) => 1);

        if(size >= BufferSize)
        {
           size = BufferSize-1;
           print('warning: clipping size'); // todo: raise error if client ask too much?
           //throw new StateError('size if too big');
        }
        
        // ajust the buffer 
        if (size != _b.length)
          _b = buffer.sublist(0,size); //todo: find a less time/memory consuming way (or not?)
         
        // send the buffer
        _socket.add(_b);
      }
    }
  }

  void onDone() {
    print('Client disconnected');
    _timer.cancel();
  }
}

void serveRequest(HttpRequest request, {int code : HttpStatus.FORBIDDEN, String reason : 'WebSocket connections only'}){
  request.response.statusCode = code;
  request.response.reasonPhrase = reason;
  request.response.close();
}

void UpgradeServer(String clientversion)
{
  //todo: 
  // 1. fetch some official version (dont trust the client!)
  // 2. if we're not in sync with official version then upgrade
  // 2bis add an option to not autoupgrade but to alert the server admin (only do this once).
  // 3. may add some timer (5mins? 10mins) to no to do this too often (bad client version spamming us => us spamming the offical source)

  // 1. for now do nothing:)
  
  // 2.
  // stop the server - if launched from launcher it will restart us, otherwise some admin script will do it 
  TheServer.close().then((e) => shutdown('update me my version is $SpeedTestVersion', code: 'RESTART'));
}
