// Copyright (c) 2014, Kirth Gersen / kirthalleg@gmail.com.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

/// common stuff between client & server
library Speedtest;
import 'dart:typed_data';
import 'dart:math';
import 'dart:collection';
import 'dart:isolate';
import 'dart:async';

// THE mighty version
// we should use a different 'protocol' version
// to distinguish changes that don't break client<->server behavior (such as UI changes)
// from changes that require synced client & server
const String SpeedTestVersion = '0.05';

// test type - enum wanted!
abstract class SpeedTestType {
  static const int PING = 1000;
  static const int DOWNLOAD = 1001;
  static const int UPLOAD = 1002;
}

const BufferSize = 10*1024*1024; // buffer size in KB, 10 MB max for now 
final Uint8List buffer = new Uint8List(BufferSize); // the buffer in bytes

// data buffer
abstract class TestBuffer {
  // init buffer 
  static void InitBuffer()
  {
    var seed = new Random();
    for (int i=0;i<buffer.length;i++)
    {
      buffer[i] = seed.nextInt(256); 
    }
  }
}

/// Connection interface
abstract class Connection {
  const Connection();
  void connect( void onConnected(String id), void onError(String id, String e) );
  void start( void onEnded(String id) );
  double get currentspeed;
  int get totalBytes;
  String get id;
  Result get result;
}

// todo: there should a result per thread so an arry of these
// todo: keep server and public ip of client in there too
// todo: give a UID to each test so we can collate threads server side
class Result {
  DateTime date;
  String server; // todo use URI type
  String server_version; 
  String client_ip; // todo use some real 'IP' type 
  int type; // SpeedTestType
  double volume = 0.0;
  double duration = 0.0; // seconds + ms
  double topspeed = 0.0;
  
  HashMap<String,String> toHashMap()
  {
     HashMap<String,String> m = new HashMap<String,String>();
     m['date'] = date.toIso8601String();
     m['server'] = server;
     m['server_version'] =  server_version;
     m['client_ip'] = client_ip;
     m['type'] = type.toString();
     m['volume'] = volume.toString();
     m['duration'] = duration.toString();
     m['topspeed'] = topspeed.toString();
     return m;
  }
  static Result fromHashMap(m)
  {
    Result r = new Result();
    r.date = DateTime.parse(m['date']);
    r.server = m['server'];
    r.server_version = m['server_version'];
    r.client_ip = m['client_ip'];
    r.type = int.parse(m['type']);
    r.volume = double.parse(m['volume']);
    r.duration = double.parse(m['duration']);
    r.topspeed = double.parse(m['topspeed']);
    return r;
  }
}

class Results {
  List<Result> connections;
  Duration duration;
  Results()
  {
    connections=new List<Result>();
    duration = Duration.ZERO;
  }
  // sum volume for same type
  double volume(int type) { 
    return connections.where((r)=>(r.type == type)).fold(0.0, (v,r)=>v+r.volume);
  }
}

class RemoteConnection implements Connection {
  String _server;
  int _type;
  var _printer; //todo: type

  Result _result;
  String _id;
  double _currentspeed=0.0;
  int _totalBytes =0;

  /// rpc stuff
  SendPort _sendPort;
  ReceivePort _receivePort;
  
  // CPS stuff
  var _onConnectedCallback;
  var _onErrorCallback;
  var _onEndedCallback;

  List<String> _args;
  Uri _uri;
  RemoteConnection(String url, this._type, this._id, this._server, int duration, int count, int msg_size, this._printer)
  {
    _receivePort = new ReceivePort();
    _receivePort.listen(_onMessage);

    _args = new List<String>();
    _args.add(_type.toString());
    _args.add(_id);
    _args.add(_server);
    _args.add(duration.toString());
    _args.add(count.toString());
    _args.add(msg_size.toString());
    _uri = Uri.parse(url);
  }
  
  Completer<Isolate> _completer = new Completer();
  Isolate _isolate;
  /// spawn the connection in an isolate
  Future<Isolate> spawn()
  {
    return Isolate.spawnUri(_uri, _args, _receivePort.sendPort).then(
        (Isolate e)
        {
          print('spawnUri completed');
          _isolate = e;
          //_isolate.errors.listen((e) {_printer(e);});
          return _completer.future;
        });
  }
  
  void _onMessage(dynamic msg)
  {
    
    if (_sendPort == null) {
      print('From isolate: got the sendport');
      _sendPort = msg;
       return;
    }
    
    print('From isolate: $msg');
    if (msg is String)
    {
      String m = msg as String;
      // valid message: C(connect),F(end) and E(error)
      String comm = m.substring(0,2);
      if (comm=='R:')
      {
        // complete the spawn here.
         _completer.complete(_isolate);
         return;
      }
      if (comm=='C:')
      {
        _onConnectedCallback(m.substring(2));
        return;
      }
      
      if (comm=='E:')
      {
        // 'E:id:message';
        List<String> mm = m.split(':');
        String id = mm[1];
        mm.removeRange(0,1);
        _onErrorCallback(id,mm.join(':'));
        return;
      }
      if (comm=='F:')
      {
        _onEndedCallback(m.substring(2));
        return;
      }
      if (comm=='M:')
      {
        _print(m.substring(2));
        return;
      }
    }
    // update for Result
    if (msg is HashMap<String,String>)
    {
      _result = Result.fromHashMap(msg);
      return;
    }
  }
  void _handleIsolate(Isolate e)
  {
    _print("isolate spawned");
  }
  
  void connect( void onConnected(String id), void onError(String id, String e) )
  {
    _onConnectedCallback = onConnected;
    _onErrorCallback = onError;
    _sendPort.send('CONNECT');
  }
  void start( void onEnded(String id) )
  {
    _onEndedCallback = onEnded;
    _sendPort.send('START');
  }
  void _print(String m)
  {
    if (_printer != null) _printer(m);
  }
  double get currentspeed => _currentspeed;
  int get totalBytes => _totalBytes;
  
  String get id => _id;
  Result get result => _result;
}